import ast
import enum
from dataclasses import dataclass

from transpiler.utils import highlight
from transpiler.exceptions import CompileError
from transpiler.phases.typing.types import TypeVariable, BaseType, TypeOperator


@dataclass
class UnresolvedTypeVariableError(CompileError):
    variable: TypeVariable

    def __str__(self) -> str:
        return f"Unresolved type variable: {self.variable}"

    def detail(self, last_node: ast.AST = None) -> str:
        if isinstance(last_node, (ast.Import, ast.ImportFrom)):
            return f"""
            This indicates the compiler was unable to infer the type of a function in a module.
            
            Currently, Typon cannot determine the type of Python functions imported from other modules, except
            for the standard library.
            As such, you need to give enough information to the compiler to infer the type of the function.
            
            For example:
                     ↓↓↓ this tells the compiler that {highlight('math.factorial')} returns an {highlight('int')}
                {highlight('res: int = math.factorial(5)')}"""
        return f"""
        This generally indicates the compiler was unable to infer the type of a variable or expression.
        A common fix is to add a type annotation to the variable or function.
        
        For example:
                     ↓↓↓ this tells the compiler that {highlight('x')} is an {highlight('int')}
            {highlight('def f(x: int):')}
        """


@dataclass
class RecursiveTypeUnificationError(CompileError):
    needle: BaseType
    haystack: BaseType

    def __str__(self) -> str:
        return f"Recursive type unification: {highlight(self.needle)} and {highlight(self.haystack)}"

    def detail(self, last_node: ast.AST = None) -> str:
        return f"""
        This generally indicates a recursive type definition. Such types are not currently supported.
        
        For example:
            {highlight('T = tuple[T]')}
            
        In the current case, {highlight(self.haystack)} contains type {highlight(self.needle)}, but an attempt was made to
        unify them.
        """


class TypeMismatchKind(enum.Enum):
    NO_COMMON_PARENT = enum.auto()
    DIFFERENT_TYPE = enum.auto()


@dataclass
class TypeMismatchError(CompileError):
    expected: BaseType
    got: BaseType
    reason: TypeMismatchKind

    def __str__(self) -> str:
        return f"Type mismatch: expected {highlight(self.expected)}, got {highlight(self.got)}"

    def detail(self, last_node: ast.AST = None) -> str:
        return f"""
        This generally indicates a type error.
        
        For example:
            {highlight('def f(x: int): ...')}
            {highlight('f("hello")')}
            
        In the current case, the compiler expected an expression of type {highlight(self.expected)}, but instead got
        an expression of type {highlight(self.got)}.
        """


@dataclass
class ArgumentCountMismatchError(CompileError):
    func: TypeOperator
    arguments: TypeOperator

    def __setattr__(self, key, value):
        print(key, value)
        super().__setattr__(key, value)

    def __str__(self) -> str:
        fcount = str(len(self.func.args))
        if self.func.variadic:
            fcount = f"at least {fcount}"

        return f"Argument count mismatch: expected {fcount}, got {len(self.arguments.args)}"

    def detail(self, last_node: ast.AST = None) -> str:
        return f"""
        This indicates missing or extraneous arguments in a function call or type instantiation.
        The called or instantiated signature was {highlight(self.func)}.
        
        Other examples:
            {highlight('def f(x: int): ...')}
            {highlight('f(1, 2)')}
        Here, the function {highlight('f')} expects one argument, but was called with two.
            
            {highlight('x: list[int, str]')}
        Here, the type {highlight('list')} expects one argument, but was instantiated with two.
        """


@dataclass
class ProtocolMismatchError(CompileError):
    value: BaseType
    protocol: BaseType
    reason: Exception

    def __str__(self) -> str:
        return f"Protocol mismatch: {highlight(self.value)} does not implement {highlight(self.protocol)}"

    def detail(self, last_node: ast.AST = None) -> str:
        return f"""
        This generally indicates a type error.
        
        For example:
            {highlight('def f(x: Iterable[int]): ...')}
            {highlight('f("hello")')}
            
        In the current case, the compiler expected an expression whose type implements {highlight(self.protocol)}, but
        instead got an expression of type {highlight(self.value)}. 
        """


@dataclass
class NotCallableError(CompileError):
    value: BaseType

    def __str__(self) -> str:
        return f"Trying to call a non-function type: {highlight(self.value)}"

    def detail(self, last_node: ast.AST = None) -> str:
        return f"""
        This indicates that an attempt was made to call an object that is not a function.
        
        For example:
            {highlight('x = 1')}
            {highlight('x()')}
        """


@dataclass
class MissingAttributeError(CompileError):
    value: BaseType
    attribute: str

    def __str__(self) -> str:
        return f"Missing attribute: {highlight(self.value)} has no attribute {highlight(self.attribute)}"

    def detail(self, last_node: ast.AST = None) -> str:
        return f"""
        This indicates that an attempt was made to access an attribute that does not exist.
        
        For example:
            {highlight('x = 1')}
            {highlight('print(x.y)')}
        """


@dataclass
class UnknownNameError(CompileError):
    name: str

    def __str__(self) -> str:
        return f"Unknown name: {highlight(self.name)}"

    def detail(self, last_node: ast.AST = None) -> str:
        return f"""
        This indicates that an attempt was made to access a name that does not exist.
        
        For example:
            {highlight('print(abcd)')}
        """


@dataclass
class UnknownModuleMemberError(CompileError):
    module: str
    name: str

    def __str__(self) -> str:
        return f"Unknown module member: Module {highlight(self.module)} does not contain {highlight(self.name)}"

    def detail(self, last_node: ast.AST = None) -> str:
        return f"""
        This indicates that an attempt was made to import 
        
        For example:
            {highlight('from math import abcd')}
        """
